@@ -33,6 +33,7 @@ from ocrd_models.ocrd_page_generateds import (
 
 from .config import OCRD_TOOL
 from ..lib.seq2seq import Sequence2Sequence, GAP
+from ..lib.alignment import Alignment
 
 TOOL_NAME = 'ocrd-cor-asv-ann-process'
 
@@ -142,8 +143,10 @@ class ANNCorrection(Processor):
                 self.logger.debug('"%s" -> "%s"', input_line.rstrip('\n'), output_line.rstrip('\n'))
                 
                 # convert soft scores (seen from output) to hard path (seen from input):
-                realignment = _alignment2path(alignment, len(input_line), len(output_line),
-                                              1. / self.s2s.voc_size)
+                #realignment = _alignment2path(alignment, len(input_line), len(output_line),
+                #                              1. / self.s2s.voc_size)
+                # create hard path via minimal edit distance:
+                realignment, distance = _alignment_path(input_line, output_line)
                 
                 # overwrite TextEquiv references:
                 new_sequence = _update_sequence(
@@ -155,8 +158,8 @@ class ANNCorrection(Processor):
                 if level != 'line':
                     _resegment_sequence(new_sequence, level)
                 
-                self.logger.info('corrected line with %d elements, ppl: %.3f',
-                                 len(new_sequence), np.exp(output_score))
+                self.logger.info('corrected line with %d elements, ppl: %.3f, CER: %.1f%%',
+                                 len(new_sequence), np.exp(output_score), distance * 100)
             
             # make higher levels consistent again:
             page_update_higher_textequiv_levels(level, pcgts)
@@ -349,12 +352,40 @@ def _alignment2path(alignment, i_max, j_max, min_score):
             j -= 1
             i -= 1
     realignment[0] = 0 # init start of line
+    # LOG = getLogger('processor.ANNCorrection')
     # LOG.debug('realignment: %s', str(realignment))
     # from matplotlib import pyplot
+    # pyplot.imshow(np.array(alignment).T)
+    # pyplot.show()
     # pyplot.imshow(viterbi_fw)
     # pyplot.show()
     return realignment
 
+def _alignment_path(input_text, output_text):
+    '''Find the minimal distance path through string pair via Smith-Waterman alignment.
+    
+    Return a dictionary mapping input positions to output positions
+    (i.e. a realignment path).
+    '''
+    alignment = Alignment.best_alignment(input_text, output_text)
+    realignment = {0: 0} # init start of line
+    i, j, dist = 0, 0, 0.0
+    for input_sym, output_sym in alignment:
+        if input_sym:
+            i += len(input_sym)
+        if output_sym:
+            j += len(output_sym)
+        if input_sym != output_sym:
+            dist += 1.0
+        realignment[i] = j
+    assert i == len(input_text)
+    assert j == len(output_text)
+    assert len(alignment) > 0
+    dist /= len(alignment) - 1 # ignore newline
+    # LOG = getLogger('processor.ANNCorrection')
+    # LOG.debug('realignment: %s', str(realignment))
+    return realignment, dist
+
 def _update_sequence(input_line, output_line, output_prob,
                      score, realignment,
                      textequivs, words, textlines):
@@ -392,6 +423,13 @@ def _update_sequence(input_line, output_line, output_prob,
         if last:
             input_ = input_line[last[0]:i]
             output = output_line[last[1]:j]
+            prob = output_prob[last[1]:j]
+            textequiv = textequivs[last[0]]
+            assert textequiv.Unicode == input_, (
+                'source element "%s" does not match input section "%s" in line "%s"' % (
+                    textequiv.Unicode, input_, line.id))
+            #print("'" + textequiv.Unicode + "' → '" + output + "'" \
+            #      "[" + str(textequiv.conf if textequiv.index != -1 else -1) + " → " + str(prob) + "]")
             # try to distribute whitespace onto whitespace, i.e.
             # if input is Whitespace, move any Non-whitespace parts
             # in output to neighbours;
@@ -399,7 +437,7 @@ def _update_sequence(input_line, output_line, output_prob,
             # if their input is Whitespace too;
             # input:  N|    W    |N   N|     W   |   W|    N    |W
             # output:  |<-N W N->|     |<-W<-N W |    |<-W N W->|
-            if input_ in (" ", "\n"):
+            if textequiv.index == -1:
                 if output and not output.startswith((" ", "\n")) and sequence:
                     while output and not output.startswith((" ", "\n")):
                         sequence[-1][0].Unicode += output[0]
@@ -423,19 +461,13 @@ def _update_sequence(input_line, output_line, output_prob,
                         last[1] += 1
                         output = output[1:]
                     #print('corrected whitespace LHS: ', last, [i, j])
-                if output.endswith((" ", "\n")) and i < i_max and input_line[i] in (" ", "\n"):
+                if output.endswith((" ", "\n")) and i < i_max and textequivs[i].index == -1:
                     while output.endswith((" ", "\n")):
                         j -= 1
                         output = output[:-1]
                     #print('corrected whitespace RHS: ', last, [i, j])
-            textequiv = textequivs[last[0]]
-            assert textequiv.Unicode == input_, (
-                'source element "%s" does not match input section "%s" in line "%s"' % (
-                    textequiv.Unicode, input_, line.id))
-            #print("'" + textequiv.Unicode + "' -> '" + output + "'")
             textequiv.Unicode = output
             #textequiv.conf = np.exp(-score)
-            prob = output_prob[last[1]:j]
             textequiv.conf = np.mean(prob or [1.0])
             word = words[last[0]]
             textline = textlines[last[0]]
@@ -445,6 +477,8 @@ def _update_sequence(input_line, output_line, output_prob,
         'alignment path did not reach top: %d/%d vs %d/%d in line "%s"' % (
             last[0], last[1], i_max, j_max, line.id))
     for i, (textequiv, _, _) in enumerate(sequence):
+        # disallow non-whitespace mapping to -1 (artificial whitespace)
+        # (but allow whitespace to anything and anything to true segments)
         assert not textequiv.Unicode.split() or textequiv.index != -1, (
             'output "%s" will be lost at (whitespace) element %d in line "%s"' % (
                 textequiv.Unicode, i, line.id))
